#---------------------------- 'db' Module ----------------------------
sub opentfile{
    my($str, $filename) = @_;
    my $path;
    my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime);
    my $language = &get_language;

    if (!$language || $language =~ /ja|jp|jis/i){
        $path = "$filename.en" unless -r ($path = "$filename.ja");
    }elsif (! -r ($path = "$filename.$language")){
        $path = $filename . '-e';
    }
    -r $path || -r ($path = $filename) || chop($path);

    if (-r $path){
        ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime)
            = stat($path) if open($str, $path);
    }
    $DbSize{$str} = $size;
    $DbTime{$str} = $mtime;
    &debug_log("opentfile: $path, $size") if $mtime;
    $mtime;
}

sub db::openzfile{
    local(*fH, $filename) = @_;
    if (-r $filename){
        if ($ZcatPri){
            $ZCatPri = 0;
            eval{setpriority(0, 0, $IniPri + $ZcatPri)};
        }
        return open(fH, "$Zcat $filename |");
    }
    return 0;
}

sub opentext{
    my ($str, $filename) = @_;
    if (!$Cgi && $Zcat && $filename =~ /\.gz$/i){
        &debug_log("opentext: gz: $filename");
        return &db::openzfile($str, $filename);
    }
    if (-r $filename){
        &debug_log("opentext: open: $filename");
        return open($str, $filename);
    }
    $filename =~ s/%([\da-f]{2})/chr hex $1/gei;
    if (-r $filename){
        &debug_log("opentext: decode: $filename");
        return open($str, $filename);
    }
    &debug_log("opentext: noexist: $filename");
    return undef;
}

sub openbfile{
    local($str, $filename, $flag) = @_;
    local(*fH) = $str;
    my($tmp) = $/;
    my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime);

    if (open(fH, $filename)){
        binmode(fH);
        ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime) = stat(fH);
        $DbSize{$str} = $size;
        $DbTime{$str} = $mtime;
        $DbNdx{$str} = $size / $IntSize if $IntSize;
        &debug_log("openbfile: $filename, $size") if $mtime;
        return $mtime;
    }

    if ($Zcat){
        my($fname);
        if (&db::openzfile(*fH, ($fname = "$filename.gz"))
            || &db::openzfile(*fH, ($fname = "$filename"."z"))){

            binmode(fH);
            undef $/;
            $fH = <fH>;
            $/ = $tmp;
            close(fH);
            if (!$?){
                ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime)
                    = stat($fname);
                $DbTime{$str} = $mtime;
                $DbSize{$str} = $size = length($fH);
                $DbNdx{$str} = $size / $IntSize if $IntSize;
                &debug_log("openzfile: $filename, $size");
                return $mtime;
            }
            $fH = '';
        }
    }
    push(@DbErrors, "(Cannot open $filename)") if $flag;
    return 0;
}

sub openfiles{
    &debug_log("openfiles: " . join(",", @_));
    my $fh = shift;
    my $ext = shift;
    local(*FH);
    my $db = '';
    my @db = ();
    my @dirlist = ('');
    for $db (@_){
        next unless defined $db;
        push(@db, $db);
        push(@dirlist, $db) if -d $db;
        push(@db, $1) if $db =~ /^\+(.+)$/;
    }
    push(@db, '') if $ext ne 'i' or !defined($_[0]);

    push(@dirlist, $BaseDir) if $BaseDir;
    push(@dirlist, $TEMPLATE) if $ext eq '';
    push(@dirlist, $DEFAULT_DIR);

    push(@dirlist, "$DiaryIndex") if -d $DiaryIndex;
    push(@dirlist, "$DiaryTemplate") if -d $DiaryTemplate;
    push(@dirlist, "$ENV{'HOME'}/Namazu") if $ENV{'HOME'};
    push(@dirlist, "$ENV{'HOME'}/Namazu/$1") if $ENV{'PWD'} =~ /^\Q$ENV{'HOME'}\E\/(Mail.*)$/;

    push(@dirlist, "/usr/local/share/namazu/index/");
    push(@dirlist, "/usr/local/var/namazu/index/");
    push(@dirlist, "/usr/local/namazu/index/");
    push(@dirlist, "/usr/var/namazu/index/");

    my $dir;
    my $dbname;

    for $dir (@dirlist){
        $dir .= '/' if $dir =~ /[^\/]$/;
        for $db (@db){
            $dbname = (length $db)? "$dir$db/$DbName": "$dir$DbName";
            &debug_log("try: $dbname");
            if ($ext){
                if ($fh? &opentfile($fh, "$dbname.$ext"):
                    -r "$dbname.$ext"){
                    my $dbdir = $dbname;
                    $dbdir =~ s/\/[^\/]+$//;
                    $Db2IdxDir{$dbname} = $dbdir;
                    $Db2IdxName{$dbname} = $db;
                    return $dbname;
                }
            }elsif ($Phone && &opentfile(*FH, "$dbname.phone")
                    or &opentfile(*FH, "$dbname.head")){
                my $line;
                $ExistMeta = 0;
                $Headname = $dbname;
                while (defined($line = <FH>)){
                    $line = &str2euc($line);
                    if (defined $ENV{'GATEWAY_INTERFACE'}){
                        if ($line =~ /^<\!DOCTYPE .* XHTML/){
                            $Xht = ' /';
                        }
                        if ($line =~ /^<\?xml .*encoding=\"([^\"]+)/
                            || $line =~
                            /\<meta.*content-type.*charset=([^\"]+)/i
                            ){
                            my $cs = $1;
                            &set_output_ja_code($cs);
                            &debug_log("meta: " . &get_output_ja_code);
                            $ExistMeta = 1;
                        }
                        if ($line =~ /\<\/head\>/i
                            && &get_language eq 'ja'
                            && !$ExistMeta){
                            &debug_log("ins_meta: " . &get_output_ja_code);
                        }
                        if ($line =~ /(\<\!-- (FILE|KEY) --\>)\s*(.*?)\s*\1/){
                            my $str = $2;
                            $DbHead{$str}{$db} = $3;
                            my @key = keys %{$DbHead{$str}};
                            if (1 < @key){
                                my $tmp = "\(";
                                my $key;
                                for $key (sort @key){
                                    $tmp .= "$key=$DbHead{$str}{$key} / ";
                                }
                                $tmp =~ s/ \/ $/\)/;
                                s/(\<\!-- $str --\>).*?\1/$1 $tmp $1/;
                            }
                        }
                    }
                    push(@HEAD, $line);
                }
                close(FH);
                @FOOT = <FH>, close(FH) if &opentfile(*FH, "$dbname.foot");
                return $dbname;
            }
        }
    }
    undef;
}

sub opendb{
    # DB ե򳫤
    my($dbname) = @_;
    my(@tmp);
    my($sec,$min,$hour,$mday,$mon,$year,$wday);

    $dbname = $DbPath if !$dbname;

    if ((-r "$dbname.access") && &openbfile(*ACCESS, "$dbname.access")){
        my($ret, $str, $op, $val);
        while (defined($str = <ACCESS>)){
            if (($op, $val) = ($str =~ /^\s*(deny|allow)\s+(\S+)/)){
                $val =~ tr/A-Z/a-z/;
                $ret = $op, next if $val eq 'all';
                if ($val =~ /[a-z]$/){
                    my $pat = quotemeta($val) . '$'; #'
                    $ret = $op if $ENV{'REMOTE_HOST'} =~ /$pat/;
                }elsif ($val =~ /^\d/){
                    my $pat = '^' . quotemeta($val);
                    $ret = $op if $ENV{'REMOTE_ADDR'} =~ /$pat/;
                }
            }
        }
        if ($ret eq 'deny'){
            push(@DbErrors,
                 "(You don\'t have a permission to access the index)");
            return;
        }
    }

    if (-e "$dbname.lock"){
        &puthtmlheader;
        if (&opentfile(*MSG, "$dbname.msg")){
            &prn($_) while <MSG>;
        }else{
            &prn("(now be in system maintenance)\n");
        }
        exit;
    }
    $DbVer2{$dbname} = $DbVer2 = &openbfile('WORDINDEX', "$dbname.wi");
    if ($DbVer2){
        &debug_log("Db: Ver2");
        return 0 unless &openbfile('WORD', "$dbname.w", 1);
    }else{
        &debug_log("Db: Ver1");
        $HashTime = &openbfile('HASH', "$dbname.h");
        &openbfile('WORD', "$dbname.w");
    }
    return 0 unless &openbfile('INDEX', "$dbname.i", 1);

    my $ht = &openbfile('INDEXINDEX', "$dbname.ii", 1);
    return 0 unless $ht;
    $HashTime = $ht unless $HashTime;

    ($sec,$min,$hour,$mday,$mon,$year,$wday) = gmtime($HashTime);
    $LastModified
        = sprintf("%s, %02d %s %04d %02d:%02d:%02d GMT",
                  ('Sun', 'Mon', 'Tue',
                   'Wed', 'Thu', 'Fri', 'Sat')[$wday], $mday,
                  ('Jan', 'Feb', 'Mar', 'Apl', 'May', 'Jun',
                   'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')[$mon],
                  $year + 1900, $hour, $min, $sec);
    &openbfile('PHRASE', "$dbname.p");
    &openbfile('PHRASEINDEX', "$dbname.pi");

    &openbfile('TIM', "$dbname.t");
    if ($SortField){
        if ($Sort =~ /^\+?fstat:/){
            if (!&openbfile('SORT', "$dbname.fstat.$SortField")
                && $FieldAlias{$SortField}){
                my $alias = $FieldAlias{$SortField};
                &openbfile('SORT', "$dbname.fstat.$alias");
            }
        }elsif (&openbfile('SORTINDEX', "$dbname.field.$SortField.i")){
            &openbfile('SORT', "$dbname.field.$SortField");
        }elsif ($FieldAlias{$SortField}){
            my $alias = $FieldAlias{$SortField};
            &openbfile('SORTINDEX', "$dbname.field.$alias.i");
            &openbfile('SORT', "$dbname.field.$alias");
        }
    }

    if ($SUMMARY and $MakeSummary || $MakeGrep){
        &openbfile("RLIST_____$DbPath", "$dbname.r");
        &openbfile("RLISTINDEX$DbPath", "$dbname.ri");
        if ($DbTime{"RLIST_____$DbPath"} > $DbTime{"RLISTINDEX$DbPath"}){
            &openbfile("RLIST_____$DbPath", "$dbname.field.uri");
            &openbfile("RLISTINDEX$DbPath", "$dbname.field.uri.i");
        }
    }
    if (!$DbVer2){
        return 0 unless &openbfile("FLIST_____$DbPath", "$dbname.f", 1);
        return 0 unless &openbfile("FLISTINDEX$DbPath", "$dbname.fi", 1);
    }
    return 1;
}

sub closefile{
    local($str) = @_;
    local(*fH) = $str;
    close(fH);
    undef $fH;
    delete $DbNdx{$str};
    delete $DbSize{$str};
    delete $DbTime{$str};
}

sub closedb{
    my($dbname) = @_;
    undef $HashTime;
    &closefile('HASH');
    &closefile('INDEX');
    &closefile('INDEXINDEX');

    &closefile('TIM');

    &closefile('PHRASE');
    &closefile('PHRASEINDEX');

    &closefile('SORTINDEX');
    &closefile('SORT');

    &closefile('WORDINDEX');
    &closefile('WORD');

    &bwclose;
    &phrase_init;

    undef %db::cache_rii_off;
    undef %db::cache_rii_buf;

    undef @SWord;
    undef @Words;
    undef %Hit;
    undef %SubHit;
    undef @ScorePtr;
    undef @DbErrors;
}

sub dbsize{
    local(*fH) = @_;
    my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size);

    return length($fH) if $fH;
    ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size) = stat(fH);
    $size;
}

sub readdb{
    local(*fH, $offset, $size) = @_;
    my($buf, $len);

    if ($fH){
        return ($offset, '') if $offset >= length($fH);
        $buf = substr($fH, $offset, $size);
    }else{
        seek(fH, $offset, 0);
        return ($offset, '') if eof(fH);
        read(fH, $buf, $size);
    }
    $offset += $size;
    ($offset, $buf);
}

sub getsdb{
    local(*fH) = shift;
    my $offset = shift;
    my($buf, $len);

    if ($fH){
        while ($offset < length($fH)){
            $c = substr($fH, $offset++, 1);
            last if $c =~ /^$/;
            $buf .= $c;
        }
    }else{
        seek(fH, $offset, 0);

        $buf = <fH>;
        $offset += length($buf);
        chomp($buf);
    }
    ($offset, $buf);
}

sub indexpointer{
    # file  N ܤ֤ͤ
    local(*fH, $n) = @_;

    $n *= $IntSize;
    my $val = &readi(*fH, $n);
    $val;
}

sub readindexindex{
    # NMZ.ii  N ܤθ NMZ.i ɤ֤
    my($x) = @_;
    my($offset, $buf);
    return ($offset, $db::cache_rii_buf{$x}) if ($offset = $db::cache_rii_off{$x});

    if ($DbVer2){
        ($offset, $buf) = &getsdb(*WORD, &indexpointer(*WORDINDEX, $x));
    }else{
        ($offset, $buf) = &getsdb(*INDEX, &indexpointer(*INDEXINDEX, $x));
    }
    $db::cache_rii_off{$x} = $offset; $db::cache_rii_buf{$x} = $buf;
    ($offset, $buf);
}

sub readindexscore{
    # NMZ.i  score ɤ߼롣ҥåȿ֤
    local($n, *score, $str) = @_;
    my($fscore, $fno, $ndx, $hit, $net, $ret);

    ($ndx, $str) = &readindexindex($n);
    $ndx = &indexpointer(*INDEXINDEX, $n) if $DbVer2;
    ($ndx, $hit) = &readw(*INDEX, $ndx);
    $hit *= $IntSize if $IntType ne 'w';
    $buf = &readdb(*INDEX, $ndx, $hit);

    my @tmp = &unpackw($buf);

    my $idf = undef;
    if ($TFIDF && $DbNdx{'TIM'}){
        $idf = log($DbNdx{'TIM'} / (scalar(@tmp)/2)) / log(2);
    }

    my $tmp;
    $fno = 0;
    while (defined($tmp = shift @tmp) && defined($fscore = shift @tmp)){
        $fscore = int($fscore * $idf) + 1 if defined $idf;

        if ($DbVer2){
            $fno += $tmp;
        }else{
            $fno = $tmp;
        }
        if (&TimEnable($fno)){
            $score{'phrase'}{$fno} .= "$str $str\n";
            $net++;
            $ret++ if !$score{$fno};
            $score{$fno} = $fscore if $fscore > $score{$fno};
        }
    }
    # ʥҥåȿʣҥåȿ֤
    ($net, $ret);
}

sub lastSize{
    my($fh, $x, $db) = @_;
    my($fhi) = $fh . 'INDEX';
    $fhi = $db if $db;
    ($x < $DbNdx{$fhi})? &indexpointer($fhi, $x): $DbSize{$fh};
}

sub lastNdx{
    my($fh, $x, $db) = @_;
    my($fhi) = $fh . 'INDEX';
    $fhi = $db if $db;
    ($x < $DbNdx{$fhi})? &indexpointer($fhi, $x): $DbNdx{$fh};
}

sub TimEnable{
    my $fileno = shift;
    my $p = $fileno;
    $p .= "#$DbPath";
    unless (defined($Tim{$p})){
        if ($fileno < $DbNdx{'TIM'}){
            $Tim{$p} = &indexpointer('TIM', $fileno);
        }else{
            $Tim{$p} = $fileno;
        }
    }
    if ($SortField){
        if ($Sort =~ /^\+?fstat:/){
            $SElem{$p} = &indexpointer('SORT', $fileno);
        }else{
            $SElem{$p} = 
                &getsdb('SORT', &indexpointer('SORTINDEX', $fileno));
        }
    }
    $Tim{$p} != $IntFF;
}

sub unpackw{
    my $x = shift;
    if ($PackWSub){
        my $ret = 0;
        my @ret = ();
        while ($x =~ s/^(.)//s){
            $ret <<= 7;
            $ret |= 0x7f & ord($1);
            push(@ret, $ret), $ret = 0 unless ord($1) & 0x80;
        }
        @ret;
    }else{
        unpack("$IntType*", $x);
    }
}

sub readw{
    local(*fH, $offset) = @_;
    return &readi(@_) if $IntType ne 'w';

    my ($ret, $c);
    if ($fH){
        while ($offset < length($fH)){
            $c = ord(substr($fH, $offset++, 1));
            $ret = ($ret << 7) | ($c & 0x7f);
            last unless $c & 0x80;
        }
    }else{
        seek(fH, $offset, 0);
        while (read(fH, $c, 1)){
            $offset++;
            $c = ord($c);
            $ret = ($ret << 7) | ($c & 0x7f);
            last unless $c & 0x80;
        }
    }
    ($offset, $ret);
}

sub readi{
    my ($offset, $ret) = &readdb($_[0], $_[1], $IntSize);
    $ret = ($ret eq '')? undef: unpack($IntNType, $ret);
    ($offset, $ret);
}

1;
#------------------------- End of 'db' Module ------------------------
