#!/usr/local/bin/perl5 -w
# -*- Perl -*-
#
# nmz-qfield.pl - program to make NMZ.qfield.*
#         by furukawa@tcp-ip.or.jp
#

use strict;
use IO::File;


my $opt_p = 0;
my $opt_v = 0;
my $debug = 0;

unless (@ARGV){
    print "Usage: perl nmz-qfield.pl [options|field|file]...\n";
    exit;
}

while (@ARGV && $ARGV[0] =~ s/^\-//){
    my $argv = shift @ARGV;
    while ($argv =~ s/^(.)//){
        my $opt = $1;
        $opt_p = 1 if $opt eq 'p';
        $opt_v = 1 if $opt eq 'v';
        $debug = 1 if $opt eq 'd';
    }
}

while (@ARGV){
    my $argv = shift @ARGV;

    my $fin = new IO::File($argv);
    next unless defined $fin;

    my $fout;
    my $fdat;
    my $fndx;

    if ($argv =~ s/NMZ\.field\.(\S+)$/NMZ/){
        $fout = new IO::File("$argv.qfield.$1", "w");
        $fdat = new IO::File("$argv.qfield.$1.f", "w");
        $fndx = new IO::File("$argv.qfield.$1.fi", "w");
    }elsif ($argv =~ /^[\.]+$/){
        $fout = new IO::File("NMZ.qfield.$argv", "w");
        $fdat = new IO::File("NMZ.qfield.$argv.f", "w");
        $fndx = new IO::File("NMZ.qfield.$argv.fi", "w");
    }else{
    }
    next unless defined($fout) && defined($fdat) && defined($fndx);

    binmode($fout);
    binmode($fdat);
    binmode($fndx);

    my %field;
    my $line = 0;
    while (defined(my $str = $fin->getline)){
        chomp $str;
        unless ($opt_p){
            $str =~ s/^[\[\(][^\(\[\)\]\x80-\xff]+\d+[\)\]]\s*//;
            $str =~ s/^\s*(Re:\s*)+//;
        }
        push(@{$field{$str}}, $line++);
    }

    my $offset = 0;
    for my $key (sort keys %field){
        my $dat = '';
        my $old = 0;

        print "$key\n" if $debug;

        for my $line (@{$field{$key}}){
            print " $line" if $debug;
            $dat .= pack('w', $line - $old);
            $old = $line;
        }
        print "\n" if $debug;

        $fout->print("$key\n");

        $fndx->print(pack('N', $offset));
        $fdat->print($dat);
        $offset += length $dat;
    }
}
